/* $Id: pbs_drmaa.c 533 2007-12-22 15:25:42Z lukasz $ */
/*
 *  FedStage DRMAA for PBS Pro
 *  Copyright (C) 2006-2007  Fedstage Systems Inc.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file pbs_drmaa.c
 * PBS specific part of DRMAA.
 */

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <pbs_error.h>

#include <drmaa_utils/attrib.h>
#include <drmaa_utils/compat.h>
#include <drmaa_utils/conf.h>
#include <drmaa_utils/drmaa_impl.h>
#include <drmaa_utils/util.h>
#include <pbs_drmaa/pbs_attrib.h>
#include <pbs_drmaa/pbs_drmaa.h>

#ifndef lint
static char rcsid[]
#	ifdef __GNUC__
		__attribute__ ((unused))
#	endif
	= "$Id: pbs_drmaa.c 533 2007-12-22 15:25:42Z lukasz $";
#endif


const drmaa_attrib_info_t pbs_drmaa_attr_table[] = {
	/* DRMAA attributes with direct PBS mapping: */
 { ATTR_JOB_NAME,         NULL, "Job_Name",       0 },
 { ATTR_ENV,              NULL, "Variable_List",  0 },
 { ATTR_OUTPUT_PATH,      NULL, "Output_Path",    0 },
 { ATTR_ERROR_PATH,       NULL, "Error_Path",     0 },
 { ATTR_JOIN_FILES,       NULL, "Join_Path",      0 },
 { ATTR_EMAIL,            NULL, "Mail_Users",     0 },
 { ATTR_START_TIME,       NULL, "Execution_Time", 0 },
 { ATTR_JOB_SUBMIT_STATE, NULL, "substate",       0 },
	/*
	 * Additional Torque/OpenPBS attribs (not in DRMAA specification;
	 * used only internaly):
	 */
 { ATTR_CHECKPOINT,        NULL, "Checkpoint"     , ATTR_F_STR },
 { ATTR_GROUP_LIST,        NULL, "group_list"     , ATTR_F_VECTOR },
 { ATTR_HOLD_TYPES,        NULL, "Hold_Types"     , ATTR_F_STR },
 { ATTR_KEEP_FILES,        NULL, "Keep_Files"     , ATTR_F_BOOL },
 { ATTR_RESOURCES,         NULL, "Resource_List"  , ATTR_F_VECTOR | ATTR_F_STR },
 { ATTR_MAIL_POINTS,       NULL, "Mail_Points"    , ATTR_F_STR },
 { ATTR_PRIORITY,          NULL, "Priority"       , ATTR_F_INT },
 { ATTR_DESTINATION_QUEUE, NULL, "destination"    , ATTR_F_STR },
 { ATTR_RERUNABLE,         NULL, "Rerunable"      , ATTR_F_BOOL },
 { ATTR_USER_LIST,         NULL, "User_List"      , ATTR_F_VECTOR | ATTR_F_STR },
 { ATTR_ACCOUNT_NAME,      NULL, "Account_Name"   , ATTR_F_STR },
 { ATTR_SHELL,             NULL, "Shell_Path_List", ATTR_F_VECTOR | ATTR_F_STR | ATTR_F_PATH },
 { ATTR_DEPENDENCIES,      NULL, "depend"         , ATTR_F_VECTOR | ATTR_F_STR },
 { ATTR_INTERACTIVE,       NULL, "interactive"    , ATTR_F_BOOL },
 { ATTR_STAGEIN,           NULL, "stagein"        , ATTR_F_VECTOR | ATTR_F_STR | ATTR_F_PATH },
 { ATTR_STAGEOUT,          NULL, "stageout"       , ATTR_F_VECTOR | ATTR_F_STR | ATTR_F_PATH },
	/* Additional job and general attribute names: */
 { ATTR_JOB_STATE,           NULL, "job_state"      , ATTR_F_STR },
 { ATTR_CTIME,               NULL, "ctime"          , ATTR_F_STR },
 { ATTR_MTIME,               NULL, "mtime"          , ATTR_F_STR },
 { ATTR_QTIME,               NULL, "qtime"          , ATTR_F_STR },
 { ATTR_ETIME,               NULL, "etime"          , ATTR_F_STR },
 { ATTR_RESOURCES_USED,      NULL, "resources_used" , 0          },
 { ATTR_EXIT_STATUS,         NULL, "exit_status"    , ATTR_F_INT },
 { ATTR_JOB_OWNER,           NULL, "Job_Owner"      , ATTR_F_STR },
 { ATTR_EUSER,               NULL, "euser"          , 0          },
 { ATTR_EGROUP,              NULL, "egroup"         , 0          },
 { ATTR_HOP_COUNT,           NULL, "hop_count"      , 0          },
 { ATTR_SECURITY,            NULL, "security"       , 0          },
 { ATTR_SCHEDULER_HINT,      NULL, "sched_hint"	    , 0          },
 { ATTR_SERVER,              NULL, "server"         , 0          },
 { ATTR_QUEUE,               NULL, "queue"          , 0          },
 { ATTR_QUEUE_RANK,          NULL, "queue_rank"     , 0          },
 { ATTR_EXECUTION_HOST,      NULL, "exec_host"      , 0          },
 { ATTR_TOTAL_JOBS,          NULL, "total_jobs"     , 0          },
 { ATTR_MAX_RUNNING,         NULL, "max_running"    , 0          },
 { ATTR_COMMENT,             NULL, "comment"        , 0          },
 { ATTR_JOB_COOKIE,          NULL, "cookie"         , 0          },
 { ATTR_HASHNAME,            NULL, "hashname"       , 0          },
 { ATTR_SESSION_ID,          NULL, "session_id"     , 0          },
 { ATTR_ALT_ID,              NULL, "alt_id"         , 0          },
 { -1, NULL, NULL, 0 } /* sentinel */
};


void
drmaa_session_create_impl( drmaa_session_t *session, drmaa_err_ctx_t *err )
{
	drmaa_session_impl_t *impl = NULL;

	if( OK(err) )
		DRMAA_MALLOC( impl, drmaa_session_impl_t );
	if( OK(err) )
		session->impl = impl;
	if( OK(err) )
	 {
		impl->pbs_conn = -1;
		impl->status_attrl = NULL;
		impl->wait_thread_started = false;
		drmaa_mutex_init( &impl->conn_mutex, err );
	 }

	if( OK(err) )
		impl->status_attrl = drmaa_create_pbs_attrl( err );

	if( OK(err) )
	 {
		drmaa_conf_dict_t *conf = session->configuration;
		const char *home = NULL;
		const char *pbs_drmaa_conf = NULL;

		conf = drmaa_conf_read( conf, "/etc/pbs_drmaa.conf", false, NULL, 0, err );
		home = getenv( "HOME" );
		if( OK(err) && home )
		 {
			char *conf_file;
			conf_file = drmaa_asprintf( err, "%s/.pbs_drmaa.conf", home );
			if( OK(err) )
				conf = drmaa_conf_read( conf, conf_file, false, NULL, 0, err );
			DRMAA_FREE( conf_file );
		 }
		pbs_drmaa_conf = getenv( "PBS_DRMAA_CONF" );
		if( OK(err) && pbs_drmaa_conf && strlen(pbs_drmaa_conf)>0 )
			conf = drmaa_conf_read( conf, pbs_drmaa_conf, true, NULL, 0, err );
		session->configuration = conf;
#ifdef DRMAA_DEBUG
		DEBUG(( "== configuration ==" ));
		drmaa_conf_dict_dump( conf );  printf( "\n" );
		DEBUG(( "== configuration ==" ));
#endif
	 }

	if( OK(err) )
	 {
		impl->pbs_conn = pbs_connect( session->contact );
		DEBUG(( "pbs_connect(%s)=%d", session->contact, impl->pbs_conn ));
		if( impl->pbs_conn < 0 )
			RAISE_PBS();
	 }

	if( OK(err) )
	 {
		drmaa_err_ctx_t q_errctx, *q_err=&q_errctx;
		struct attrl default_queue_query;
		struct attrl keep_completed_query;
		struct batch_status *default_queue_result = NULL;
		struct batch_status *keep_completed_result = NULL;
		/* char *default_queue = NULL; */
		default_queue_query.next = NULL;
		default_queue_query.name = "default_queue";
		default_queue_query.resource = NULL;
		default_queue_query.value = NULL;
		keep_completed_query.next = NULL;
		keep_completed_query.name = "keep_completed";
		keep_completed_query.resource = NULL;
		keep_completed_query.value = NULL;

		session->missing_jobs = DRMAA_IGNORE_MISSING_JOBS;
		drmaa_err_create( q_err, err );

		/* XXX: crashes with OpenPBS library. */
#if 0
		if( OK(q_err) )
		 {
			default_queue_result =
					pbs_statserver( impl->pbs_conn, &default_queue_query, NULL );
			if( default_queue_result == NULL )
				drmaa_err_pbs_error( q_err );
			else
			 {
				if( default_queue_result->attribs
						&&  !strcmp( default_queue_result->attribs->name,
							"default_queue" ) )
					default_queue = default_queue_result->attribs->value;
				else
					drmaa_err_drmaa_error(
							q_err, DRMAA_ERRNO_INTERNAL_ERROR,
							"no default queue set for PBS server"
							);
			 }
		 }

		if( OK(q_err) )
		 {
			DEBUG(( "default_queue: %s", default_queue ));
			keep_completed_result = pbs_statque( impl->pbs_conn, default_queue,
					&keep_completed_query, NULL );
			if( keep_completed_result == NULL )
			 {
				drmaa_err_pbs_error( q_err );
			 }
			else
			 {
				if( keep_completed_result->attribs
						&&  !strcmp( keep_completed_result->attribs->name,
							"keep_completed_result" ) )
				 {
					DEBUG(( "keep_completed: %s",
								keep_completed_result->attribs->value ));
					session->missing_jobs = DRMAA_IGNORE_QUEUED_MISSING_JOBS;
				 }
				else
					DEBUG(( "keep_completed is not set for default queue" ));
			 }
		 }
#endif

		if( default_queue_result )
			pbs_statfree( default_queue_result );
		if( keep_completed_result )
			pbs_statfree( keep_completed_result );

		drmaa_err_destroy( q_err );
	 }

	if( OK(err) )
		drmaa_session_apply_config( session, err );

	if( OK(err)  &&  session->with_wait_thread )
	 {
		drmaa_thread_create( &impl->wait_thread, drmaa_wait_thread, session, err );
		if( OK(err) )
			impl->wait_thread_started = true;
	 }

	if( OK(err) )
	 {
		const drmaa_attrib_info_t *i;
		for( i = pbs_drmaa_attr_table;  i->code >= 0;  i++ )
			drmaa_register_attribute( session, i->code, NULL, i->drm_name, i->flags, err );
	 }
}


void
drmaa_session_destroy_impl( drmaa_session_t *session, drmaa_err_ctx_t *err )
{
	drmaa_session_impl_t *impl;

	impl = session->impl;
	if( impl == NULL )
		return;

	if( impl->wait_thread_started )
	 {
		drmaa_thread_join( impl->wait_thread, NULL, err );
		impl->wait_thread_started = false;
	 }

	if( impl->status_attrl != NULL )
		DRMAA_FREE( impl->status_attrl );
	if( impl->pbs_conn >= 0 )
	 {
		int rc = 0;
		rc = pbs_disconnect( impl->pbs_conn );
		if( rc )
			RAISE_PBS();
	 }
	drmaa_mutex_destroy( &impl->conn_mutex, err );
	DRMAA_FREE( impl );
	session->impl = NULL;
}



void
drmaa_control_impl(
		drmaa_session_t *session, drmaa_job_t *job, int action,
		drmaa_err_ctx_t *err
		)
{
	drmaa_session_impl_t *impl = NULL;
	const char *job_id = NULL;
	int rc = -1;
	int i;

	if( !OK(err) )
		return;
	DEBUG(( "-> drmaa_control_impl( job_id=%s, action=%s )",
				job->job_id, drmaa_control_to_str(action) ));
	impl = session->impl;
	job_id = job->job_id;

	/*
	 * In PBS Pro pbs_sigjob raises internal server error (PBSE_INTERNAL)
	 * when job just changed its state to running.
	 */
	for( i=0;  rc!=0 && i<3;  i++ )
	 {
		drmaa_mutex_lock( &impl->conn_mutex, err );
		switch( action )
		 {
			/*
			 * We cannot know whether we did suspend job
			 * in other way than remembering this inside DRMAA session.
			 */
			case DRMAA_CONTROL_SUSPEND:
				rc = pbs_sigjob( impl->pbs_conn, (char*)job_id, "SIGSTOP", NULL );
				if( rc == 0 )
					job->flags |= DRMAA_JOB_SUSPENDED;
				break;
			case DRMAA_CONTROL_RESUME:
				rc = pbs_sigjob( impl->pbs_conn, (char*)job_id, "SIGCONT", NULL );
				if( rc == 0 )
					job->flags &= ~DRMAA_JOB_SUSPENDED;
				break;
			case DRMAA_CONTROL_HOLD:
				rc = pbs_holdjob( impl->pbs_conn, (char*)job_id, USER_HOLD, NULL );
				if( rc == 0 )
					job->flags |= DRMAA_JOB_HOLD;
				break;
			case DRMAA_CONTROL_RELEASE:
				rc = pbs_rlsjob( impl->pbs_conn, (char*)job_id, USER_HOLD, NULL );
				if( rc == 0 )
					job->flags &= DRMAA_JOB_HOLD;
				break;
			case DRMAA_CONTROL_TERMINATE:
				rc = pbs_deljob( impl->pbs_conn, (char*)job_id, NULL );
				  /* deldelay=N -- delay between SIGTERM and SIGKILL (default 0) */
				if( rc == 0 )
				 {
					job->flags &= DRMAA_JOB_TERMINATED_MASK;
					if( (job->flags & DRMAA_JOB_TERMINATED) == 0 )
						job->flags |= DRMAA_JOB_TERMINATED | DRMAA_JOB_ABORTED;
				 }
				break;
		 }
		drmaa_mutex_unlock( &impl->conn_mutex, err );

		if( rc == PBSE_INTERNAL )
		 {
			DEBUG(( "drmaa_control_impl: repeating request (%d of 3)", i+2 ));
			sleep( 1 );
		 }
	 }

	if( rc )
		RAISE_PBS();
	DEBUG(( "<- drmaa_control_impl; err=%d", err->rc ));
}


void
drmaa_job_update_status_impl( drmaa_job_t *job, struct attrl *attribs,
		drmaa_err_ctx_t *err );


struct attrl *
drmaa_create_pbs_attrl( drmaa_err_ctx_t *err )
{
	struct attrl *result = NULL;
	struct attrl *i;
	const int max_attribs = 16;
	int n_attribs;

	DEBUG(( "-> drmaa_create_pbs_attrl" ));
	DRMAA_CALLOC( result, max_attribs, struct attrl );
	if( OK(err) )
	 {
		int j = 0;
		result[j++].name="job_state";
		result[j++].name="exit_status";
		result[j++].name="resources_used";
		result[j++].name="ctime";
		result[j++].name="mtime";
		result[j++].name="qtime";
		result[j++].name="etime";
		n_attribs = j;
#if 0
		result[j].name="resources_used";  result[j].resource="walltime";  j++;
		result[j].name="resources_used";  result[j].resource="cput";  j++;
		result[j].name="resources_used";  result[j].resource="mem";  j++;
		result[j].name="resources_used";  result[j].resource="vmem";  j++;
		result[j].name="Resource_List";  result[j].resource="walltime";  j++;
		result[j].name="Resource_List";  result[j].resource="cput";  j++;
		result[j].name="Resource_List";  result[j].resource="mem";  j++;
		result[j].name="Resource_List";  result[j].resource="vmem";  j++;
#endif
		for( i = result;  true;  i++ )
			if( i+1 < result + n_attribs )
				i->next = i+1;
			else
			 {
				i->next = NULL;
				break;
			 }
	 }

#ifdef DRMAA_DEBUG
	DEBUG(( "<- drmaa_create_pbs_attrl:" ));
	drmaa_log_attrl( result, NULL );
#endif
	return result;
}


void
drmaa_update_all_jobs_status( drmaa_session_t *c, drmaa_err_ctx_t *err )
{
	struct batch_status *status = NULL;
	struct batch_status *i;
	DEBUG(( "-> drmaa_update_all_jobs_status" ));

	if( OK(err) )
	 {
		drmaa_mutex_lock( &c->impl->conn_mutex, err );
		/* XXX: simple workaround to work with OpenPBS */
#		if 1
		status = pbs_statjob( c->impl->pbs_conn, NULL, NULL, NULL );
#		else
		status = pbs_statjob( c->impl->pbs_conn, NULL, c->impl->status_attrl, NULL );
#		endif
		DEBUG(( "pbs_statjob( fd=%d, job_id=NULL, attribs={...} ) =%p",
				 c->impl->pbs_conn, (void*)status ));
		if( status == NULL  &&  pbs_errno != 0 )
			RAISE_PBS();
		drmaa_mutex_unlock( &c->impl->conn_mutex, err );
	 }

	if( OK(err) )
	 {
		drmaa_job_set_t *jobs = c->jobs;
		drmaa_job_t *job;
		size_t i;
		drmaa_mutex_lock( &jobs->mutex, err );
		for( i = 0;  i < jobs->tab_size;  i++ )
			for( job = jobs->tab[i];  job != NULL;  job = job->next )
			 {
				drmaa_mutex_lock( &job->mutex, err );
				if( !(job->flags & DRMAA_JOB_DISPOSED) )
					job->flags |= DRMAA_JOB_MISSING;
				drmaa_mutex_unlock( &job->mutex, err );
			 }
		drmaa_mutex_unlock( &jobs->mutex, err );
	 }

	for( i = status;  OK(err) && i != NULL;  i = i->next )
	 {
		drmaa_job_t *job = NULL;
		DEBUG(( "job_id=%s", i->name ));
		if( OK(err) )
		 {
			job = drmaa_job_find( c->jobs, i->name, err );
			if( err->rc == DRMAA_ERRNO_INVALID_JOB )
				drmaa_err_clear( err );
		 }
		if( OK(err)  &&  job )
		 {
			job->flags &= ~DRMAA_JOB_MISSING;
			drmaa_job_update_status_impl( job, i->attribs, err );
		 }

		if( job )
			drmaa_job_release( &job, err );
	 }

	if( OK(err) )
	 {
		drmaa_job_set_t *jobs = c->jobs;
		drmaa_job_t *job;
		size_t i;
		drmaa_mutex_lock( &jobs->mutex, err );
		for( i = 0;  i < jobs->tab_size;  i++ )
			for( job = jobs->tab[i];  job != NULL;  job = job->next )
			 {
				drmaa_mutex_lock( &job->mutex, err );
				if( job->flags & DRMAA_JOB_MISSING )
					drmaa_job_missing( job, err );
				drmaa_mutex_unlock( &job->mutex, err );
			 }
		drmaa_mutex_unlock( &jobs->mutex, err );
	 }


	if( status != NULL )
		pbs_statfree( status );

	DEBUG(( "<- drmaa_update_all_jobs_status; err=%d", err->rc ));
}


void
drmaa_job_update_status( drmaa_job_t *job, drmaa_err_ctx_t *err )
{
	drmaa_session_t *c = job->session;
	struct batch_status *status = NULL;

	DEBUG(( "-> drmaa_job_update_status( job_id=%s )", job->job_id ));
	if( OK(err) )
	 {
		drmaa_mutex_lock( &c->impl->conn_mutex, err );
		/* XXX: simple workaround to work with OpenPBS */
#		if 1
		status = pbs_statjob( c->impl->pbs_conn, job->job_id, NULL, NULL );
#		else
		status = pbs_statjob( c->impl->pbs_conn, job->job_id, c->impl->status_attrl, NULL );
#		endif
		DEBUG(( "pbs_statjob( fd=%d, job_id=%s, attribs={...} ) =%p",
				 c->impl->pbs_conn, job->job_id, (void*)status ));
		if( status == NULL )
			switch( pbs_errno )
			 {
				case PBSE_UNKJOBID:
					break;
				default:
					RAISE_PBS();
					break;
				case 0:  /* ? */
					RAISE_DRMAA_1( DRMAA_ERRNO_INTERNAL_ERROR );
					break;
			 }
		drmaa_mutex_unlock( &c->impl->conn_mutex, err );
	 }

	if( OK(err) )
	 {
		if( status != NULL )
		 {
			/* assert( !strcmp( job->job_id, status->name ) ); */
			drmaa_job_update_status_impl( job, status->attribs, err );
		 }
		else
			drmaa_job_missing( job, err );
	 }

	if( status != NULL )
		pbs_statfree( status );

	DEBUG(( "<- drmaa_job_update_status; err=%d", err->rc ));
}


void
drmaa_job_update_status_impl( drmaa_job_t *job, struct attrl *attribs,
		drmaa_err_ctx_t *err )
{
	struct attrl *i = NULL;
	char pbs_state = 0;
	int exit_status = -1;
	const char *cpu_usage = NULL;
	const char *mem_usage = NULL;
	const char *walltime = NULL;

	DEBUG(( "-> drmaa_job_update_status_impl( job_id=%s )", job->job_id ));
#ifdef DRMAA_DEBUG
	drmaa_log_attrl( attribs, NULL );
#endif

	for( i = attribs;  OK(err) && i != NULL;  i = i->next )
	 {
		const drmaa_attrib_info_t *attr;
		attr = attr_by_drm_name( job->session, i->name, err );
		if( attr )
			switch( attr->code )
			 {
				case ATTR_JOB_STATE:
					pbs_state = i->value[0];
					break;
				case ATTR_EXIT_STATUS:
					exit_status = atoi( i->value );
					break;
				case ATTR_RESOURCES_USED:
					if( !strcmp( i->resource, "cput" ) )
						cpu_usage = i->value;
					else if( !strcmp( i->resource, "mem" ) )
						mem_usage = i->value;
					else if( !strcmp( i->resource, "walltime" ) )
						walltime = i->value;
					break;
			 }
	 }

	if( pbs_state )
		DEBUG(( "pbs_state: %c", pbs_state ));

	if( exit_status != -1 )
	 {
		DEBUG(( "exit_status: %d", exit_status ));
		job->exit_status = exit_status;
	 }

	switch( pbs_state )
	 {
		case 'C': /* Job is completed after having run. */
		case 'E': /* Job is exiting after having run. */
			job->flags &= DRMAA_JOB_TERMINATED_MASK;
			job->flags |= DRMAA_JOB_TERMINATED;
			if( job->exit_status == 0 )
				job->status = DRMAA_PS_DONE;
			else
				job->status = DRMAA_PS_FAILED;
			break;
		case 'H': /* Job is held. */
			job->status = DRMAA_PS_USER_ON_HOLD;
			job->flags |= DRMAA_JOB_HOLD;
			break;
		case 'Q': /* Job is queued, eligible to run or routed. */
		case 'W': /* Job is waiting for its execution time to be reached. */
			job->status = DRMAA_PS_QUEUED_ACTIVE;
			job->flags &= ~DRMAA_JOB_HOLD;
			break;
		case 'R': /* Job is running. */
		case 'T': /* Job is being moved to new location (?). */
		 {
			if( job->flags & DRMAA_JOB_SUSPENDED )
				job->status = DRMAA_PS_USER_SUSPENDED;
			else
				job->status = DRMAA_PS_RUNNING;
			break;
		 }
		case 'S': /* (Unicos only) job is suspend. */
			job->status = DRMAA_PS_SYSTEM_SUSPENDED;
			break;
		case 0:  default:
			job->status = DRMAA_PS_UNDETERMINED;
			break;
	 }

	if( OK(err) )
		DEBUG(( "job_ps: %s", drmaa_job_ps_to_str(job->status) ));

	if( OK(err) )
	 {
		int hours, minutes, seconds;
		long mem;
		if( cpu_usage && sscanf( cpu_usage, "%d:%d:%d",
				&hours, &minutes, &seconds ) == 3 )
		 {
			job->cpu_usage = 60*( 60*hours + minutes ) + seconds;
			DEBUG(( "cpu_usage: %s=%lds", cpu_usage, job->cpu_usage ));
		 }
		if( mem_usage && sscanf( mem_usage, "%ldkb", &mem ) == 1 )
		 {
			job->mem_usage = 1024*mem;
			DEBUG(( "mem_usage: %s=%ldB", mem_usage, job->mem_usage ));
		 }
		if( walltime && sscanf( walltime, "%d:%d:%d",
					&hours, &minutes, &seconds ) == 3 )
		 {
			job->walltime = 60*( 60*hours + minutes ) + seconds;
			DEBUG(( "walltime: %s=%lds", walltime, job->walltime ));
		 }
	 }
}


void
drmaa_job_missing( drmaa_job_t *job, drmaa_err_ctx_t *err )
{
	unsigned missing_mask = 0;

	DEBUG(( "-> drmaa_job_missing( job_id=%s )", job->job_id ));
	switch( job->session->missing_jobs )
	 {
		case DRMAA_RAISE_MISSING_JOBS:          missing_mask = 0;     break;
		case DRMAA_IGNORE_MISSING_JOBS:         missing_mask = 0x73;  break;
		case DRMAA_IGNORE_QUEUED_MISSING_JOBS:  missing_mask = 0x13;  break;
	 }
	DEBUG(( "last job_ps: %s (0x%02x); mask: 0x%02x",
				drmaa_job_ps_to_str(job->status), job->status, missing_mask ));

	if( job->status < DRMAA_PS_DONE
			&&  (job->status & ~missing_mask) )
		RAISE_DRMAA_2(
				DRMAA_ERRNO_INTERNAL_ERROR,
				"job missing from queue"
				);

	if( (job->status & DRMAA_JOB_TERMINATED_MASK) == 0 )
	 {
		job->flags &= DRMAA_JOB_TERMINATED_MASK;
		job->flags |= DRMAA_JOB_TERMINATED;
	 }

	if( (job->flags & DRMAA_JOB_ABORTED) == 0
			&&  job->session->missing_jobs == DRMAA_IGNORE_MISSING_JOBS )
	 { /* assume everthing was ok */
		job->status = DRMAA_PS_DONE;
		job->exit_status = 0;
	 }
	else
	 { /* job aborted */
		job->status = DRMAA_PS_FAILED;
		job->exit_status = 256 - SIGTERM;
	 }

	DEBUG(( "<- drmaa_job_missing; err=%d, job_ps=%s, exit_status=%d",
				err->rc, drmaa_job_ps_to_str(job->status), job->exit_status ));
}


const drmaa_implementation_info_t *
drmaa_get_implementation_info( drmaa_err_ctx_t *err )
{
	static const char *const attributes[] = {
		"drmaa_job_name",
		"drmaa_remote_command",
		"drmaa_input_path",
		"drmaa_output_path",
		"drmaa_error_path",
		"drmaa_join_files",
		"drmaa_transfer_files",
		"drmaa_wd",
		"drmaa_block_email",
		"drmaa_start_time",
		"drmaa_js_state",
		"drmaa_duration_hlimit",
		"drmaa_wct_hlimit",
		"drmaa_job_category",
		"drmaa_native_specification",
		NULL
	};
	static const char *const vector_attributes[] = {
		"drmaa_v_argv",
		"drmaa_v_env",
		"drmaa_v_email",
		NULL
	};
	static drmaa_implementation_info_t drmaa_info = {
		/* drm_system: */ "PBS/Torque",
		/* default_contact: */ NULL,
		/* drmaa_implementation: */
			PACKAGE_NAME" v. "PACKAGE_VERSION
					" <http://sourceforge.net/projects/pbspro-drmaa/>",
		/* attributes: */ attributes,
		/* vector attributes: */ vector_attributes
	};
	drmaa_info.default_contact = pbs_default();
	if( drmaa_info.default_contact == NULL )
		drmaa_info.default_contact = "localhost";
	return &drmaa_info;
}


